{-# LANGUAGE LambdaCase #-}
{-# LANGUAGE OverloadedRecordDot #-}
{-# LANGUAGE NPlusKPatterns #-}
{-# LANGUAGE MultiWayIf #-}
module Graphics.Text.Box (Box, square, Border(..), asciiBorder, grid, asciiGrid, size, lookup, render) where

import Prelude hiding (lookup)

import Data.Array.IArray (Array, genArray)
import Data.Text (Text)
import qualified Data.Array.IArray as Array
import Control.Monad (join)
import Data.Bifunctor ( Bifunctor(bimap) )
import qualified Data.Maybe as Maybe
import qualified Data.Foldable as Foldable
import Control.Arrow (Arrow ((***)))
import qualified Data.Text as Text

-- | A box of text. Has rectangular size, queried with 'size'.

data Box
  = Raw (Array (Int, Int) Char)
  | Grid (Array (Int, Int) Box) (Maybe Border)

data Border = Border
  { top, right, left, bottom, topRight, bottomRight, bottomLeft, topLeft, crosssection, verticalBetween, horizontalBetween :: Char }

-- | Draw grids like this:
--
-- @
-- \/---\
-- | | |
-- |-+-|
-- | | |
-- \\---/
-- @

asciiBorder :: Border
asciiBorder = Border {top='-', right='|', left='|', bottom='-', topRight='\\', bottomRight='/', bottomLeft='\\', topLeft='/', crosssection='+', verticalBetween='|', horizontalBetween='-'}

type Row = Int
type Column = Int

square :: Int -> ((Row, Column) -> Char) -> Box
square dim f = Raw $ genArray ((0, 0), (dim - 1, dim - 1)) f

-- | All fields will be laid out like in a static grid. There is no compensation for differently sized Boxes in different rows. 

grid :: Maybe Border -> (Int, Int) -> ((Int, Int) -> Box) -> Box
grid border (rows, cols) f = let

  boxes = genArray ((0, 0), (rows - 1, cols - 1)) f

  in Grid boxes border

-- | Use the 'asciiBorder'.

asciiGrid :: (Int, Int) -> ((Int, Int) -> Box) -> Box
asciiGrid = grid $ Just asciiBorder

gridCellSize :: Array (Int, Int) Box -> (Int, Int)
gridCellSize = let
  updateMaxSize box (height, width) = (max height *** max width) $ size box
  in Foldable.foldr updateMaxSize (0, 0) 

-- | The rectangular size of the box.

size :: Box -> (Int, Int)
size = \case
  Raw chars -> arraySize chars -- snd takes the higher bound, +1 because the array has inclusive ranges
  Grid boxes border -> let

    (boxRows, boxColumns) = arraySize boxes
    (cellHeight, cellWidth) = gridCellSize boxes
    borderSize = case border of
      Just _  -> 1
      Nothing -> 0

    in (boxRows * (cellHeight + borderSize) + 1 , boxColumns * (cellWidth + borderSize) + 1)

arraySize :: Array (Int, Int) a -> (Int, Int)
arraySize = join bimap (+1) . snd . Array.bounds

orElse :: Maybe a -> a -> a
orElse option x = Maybe.fromMaybe x option

-- | Look at a single character in the box.

lookup :: Box -> (Int, Int) -> Maybe Char
lookup b index@(row, col) = case b of
  Raw chars -> chars Array.!? index
  Grid boxes borderM -> let

     (boxHeight, boxWidth) = gridCellSize boxes
     (totalHeight, totalWidth) = size b

     in case borderM of

      Nothing -> let
        (boxRow, boxRowIndex) = row `divMod` boxHeight
        (boxCol, boxColIndex) = col `divMod` boxWidth
        in do
          innerBox <- boxes Array.!? (boxRow, boxCol)
          lookup innerBox (boxRowIndex, boxColIndex)

      Just border -> let
        (boxRow, boxRowIndex) = pred row `divMod` (boxHeight + 1)
        (boxCol, boxColIndex) = pred col `divMod` (boxWidth + 1)

        in if
          | row == 0 && col == 0                                -> Just border.topLeft
          | col == 0 && row == totalHeight - 1                  -> Just border.bottomLeft
          | row == 0 && col == totalWidth  - 1                  -> Just border.topRight
          | row == totalHeight - 1 && col == totalWidth - 1     -> Just border.bottomRight
          | row == 0                                            -> Just border.top
          | row == totalHeight - 1                              -> Just border.bottom
          | col == 0                                            -> Just border.left
          | col == totalWidth - 1                               -> Just border.right
          | boxRowIndex == boxHeight && boxColIndex == boxWidth -> Just border.crosssection
          | boxRowIndex == boxHeight                            -> Just border.horizontalBetween
          | boxColIndex == boxWidth                             -> Just border.verticalBetween
          | otherwise -> do
            innerBox <- boxes Array.!? (boxRow, boxCol)
            lookup innerBox (boxRowIndex, boxColIndex)

-- | Render everything inside, recursively. Produces a rectangular string with line-breaks.

render :: Box -> Text
render box = let
    (rows, cols) = join bimap pred $ size box
    in Text.unlines
      [ 
        Text.pack [
          lookup box (row, col) `orElse` ' '
        | col <- [0..cols]
        ]
      | row <- [0..rows]
      ]


