-- | A t'Sudoku' is a 3x3 grid of 3x3 grids.

{-# LANGUAGE GeneralizedNewtypeDeriving #-}
{-# LANGUAGE UndecidableInstances #-}
{-# LANGUAGE DerivingVia #-}
{-# LANGUAGE InstanceSigs #-}
{-# LANGUAGE DeriveTraversable #-}
{-# LANGUAGE TypeApplications #-}
module Sudoku (Sudoku(..), FieldIndex, parse, entireRowOf, entireColumnOf, entireAreaOf, translateIntIndex, orderedIntIndices, orderedIndices, affectedFieldsOf) where

import Sudoku.Area (Area)
import Data.Functor.Compose (Compose)
import Text.Megaparsec (Parsec)
import Data.Void (Void)
import Sudoku.Field (Field)
import Data.Text (Text)
import Data.Array.IArray (Array)
import Data.Array (listArray)
import qualified Data.List as List
import qualified Sudoku.Field as Field
import qualified Text.Megaparsec.Char as Megaparsec
import Control.Monad (replicateM, join)
import Control.Monad.Representable.Reader
import Data.Distributive (Distributive (..))
import qualified Data.Array.IArray as Array
import Sudoku.Triple (Index)
import Data.Enum.Util (enumerate)
import Data.Bifunctor (bimap)
import qualified Data.Set as Set
import Data.Set (Set)
import qualified Data.Functor.Rep as Representable

-- | Composing two areas produces a Sudoku.
--
-- Index order is:
--   - outer row
--   - outer column
--   - inner row
--   - inner column

type FieldIndex = ((Index, Index), (Index, Index))

newtype Sudoku a = Sudoku { get :: Compose Area Area a }
  deriving stock (Foldable, Traversable)
  deriving newtype (Functor, Representable, Show)

instance Distributive Sudoku where
  distribute :: Functor f => f (Sudoku a) -> Sudoku (f a)
  distribute = Sudoku . distribute . fmap get

parseSudokuArray :: Parsec Void Text (Array (Int, Int) (Maybe Field))
parseSudokuArray = listArray ((0, 0), (8, 8)) . List.concat <$> (9 `times` sudokuLine)

sudokuLine :: Parsec Void Text [Maybe Field]
sudokuLine = (9 `times` Field.parse) <* Megaparsec.eol

times :: Monad m => Int -> m a -> m [a]
times = replicateM

-- | Numbers are read from digits, 0 is for empty fields.

parse :: Parsec Void Text (Sudoku (Maybe Field))
parse = do
  numbers <- parseSudokuArray

  let lookupIndex ix = numbers Array.! translateIndex ix

  pure (tabulate lookupIndex)

translateIndex :: FieldIndex -> (Int, Int)
translateIndex ((outerRow, outerCol), (innerRow, innerCol)) = let
        translatePair major minor = fromEnum major * 3 + fromEnum minor
        row = translatePair outerRow innerRow
        col = translatePair outerCol innerCol
        in (row, col)

translateIntIndex :: (Int, Int) -> FieldIndex
translateIntIndex (row, col) = let

        (outerRow, innerRow) = join bimap toEnum $ row `divMod` 3
        (outerCol, innerCol) = join bimap toEnum $ col `divMod` 3

        in ((outerRow, outerCol), (innerRow, innerCol))

orderedIntIndices :: (Enum a, Num a) => [[(a, a)]]
orderedIntIndices = [ [(i, j) | j <- [0..8]] | i <- [0..8]]

orderedIndices :: [[FieldIndex]]
orderedIndices = (translateIntIndex <$>) <$> orderedIntIndices

-- >>> import Data.Functor.Rep
-- >>> :kind! Rep Sudoku
-- Rep Sudoku :: *
-- = ((Index, Index), (Index, Index))

-- | Calculate the indices of all the fields in the same row as the argument.

entireRowOf :: FieldIndex -> [FieldIndex]
entireRowOf ((_, outerColumn), (_, innerColumn)) = [((outerRow, outerColumn), (innerRow, innerColumn)) | outerRow <- enumerate, innerRow <- enumerate]

-- | Calculate the indices of all the fields in the same column as the argument.

entireColumnOf :: FieldIndex -> [FieldIndex]
entireColumnOf ((outerRow, _), (innerRow, _)) = do
  outerColumn <- enumerate
  innerColumn <- enumerate
  pure ((outerRow, outerColumn), (innerRow, innerColumn))

-- | Calculate all the indices of the fields in the same area as the argument.

entireAreaOf :: FieldIndex -> [FieldIndex]
entireAreaOf (outer, _) = do
  innerRow <- enumerate
  innerColumn <- enumerate
  pure (outer, (innerRow, innerColumn))

affectedFieldsOf :: FieldIndex -> Set FieldIndex
affectedFieldsOf = Representable.index affectedFields

affectedFields :: Sudoku (Set FieldIndex)
affectedFields = tabulate @Sudoku (Set.fromList . allAffectedFieldsOf)
  where
    allAffectedFieldsOf :: FieldIndex -> [FieldIndex]
    allAffectedFieldsOf position = List.concatMap ($ position) [entireRowOf, entireColumnOf, entireAreaOf]

