{-# LANGUAGE NoFieldSelectors #-}
{-# LANGUAGE InstanceSigs #-}
{-# LANGUAGE LambdaCase #-}
{-# LANGUAGE TupleSections #-}
module Sudoku.State (SudokuState(..), State) where

import Control.Arrow ((>>>))

import Data.Bifunctor (Bifunctor(bimap))
import Data.Enum.Util (enumerate)
import Data.Functor.Rep (imapRep)
import Data.Set (Set)
import Data.Map (Map)
import Data.Maybe (mapMaybe, catMaybes, fromMaybe)

import qualified Data.Foldable as Foldable
import qualified Data.Functor.Rep as Representable
import qualified Data.Functor.Representable as Representable
import qualified Data.Map as Map
import qualified Data.Maybe as Maybe
import qualified Data.Set as Set

import Sudoku (Sudoku, FieldIndex, entireAreaOf, entireRowOf, entireColumnOf, affectedFieldsOf)
import Sudoku.Field (Field)

class SudokuState a where
  build :: Sudoku (Maybe Field) -> a
  isValid :: a -> Bool
  nextField :: a -> Maybe FieldIndex
  fieldOptions :: a -> FieldIndex -> Set Field
  setField :: a -> FieldIndex -> Field -> a

data State = State
  { fieldOptions :: Sudoku (Maybe (Set Field))
  , optionCounts :: Map Int (Set FieldIndex)
  }
  deriving (Show)

collectConstraints :: Sudoku (Maybe Field) -> FieldIndex -> Maybe Field -> Maybe (Set Field)
collectConstraints sudoku position = \case
      Just _ -> Nothing
      Nothing -> Just $ let
        collectPosition   = Set.fromList . mapMaybe (Representable.index sudoku)

        areaConstraints   = collectPosition $ Sudoku.entireAreaOf position
        rowConstraints    = collectPosition $ Sudoku.entireRowOf position
        columnConstraints = collectPosition $ Sudoku.entireColumnOf position

        in Set.unions [areaConstraints, rowConstraints, columnConstraints]

collectOptions :: Sudoku (Maybe Field) -> FieldIndex -> Maybe Field -> Maybe (Set Field)
collectOptions sudoku position fieldValue = (Set.fromList enumerate `Set.difference`) <$> collectConstraints sudoku position fieldValue

instance SudokuState State where
  build :: Sudoku (Maybe Field) -> State
  build sudoku = let

    options = imapRep (collectOptions sudoku) sudoku
    counts  = imapRep (\ pos -> fmap (, pos)) -- Sudoku (Maybe (FieldIndex, Set Field))
      >>> catMaybes . Foldable.toList -- keep only non-present types
      >>> fmap (bimap Set.size Set.singleton) -- set keys and values up
      >>> Map.fromListWith Set.union -- combine them all
      $ options

    in State options counts

  isValid :: State -> Bool
  isValid (State options counts) = let
    optionsValid = all (maybe True (not . Set.null)) options
    countsValid = Maybe.isNothing (Map.lookup 0 counts)

    in optionsValid && countsValid

  nextField :: State -> Maybe FieldIndex
  nextField (State _ counts) = Map.lookupMin counts >>= Set.lookupMin . snd

  fieldOptions :: State -> FieldIndex -> Set Field
  fieldOptions (State options _) = fromMaybe Set.empty . Representable.index options

  setField :: State -> FieldIndex -> Field -> State
  setField (State options counts) position value = let

    affectedFields = affectedFieldsOf position

    newOptions = Representable.set position Nothing     -- remove the field where the value was set
      >>> adjustAffectedOptions affectedFields value    -- remove a lot of options in affected fields
      $ options

    oldOptionsCount = maybe 0 Set.size $ Representable.index options position -- how many options did the field have before it was set?
    newCounts = Map.adjust (Set.delete position) oldOptionsCount -- forget the set field
      >>> adjustAffectedCounts options newOptions affectedFields
      >>> Map.filter (not . Set.null) -- remove empty sets, they would interfere with future operations
      $ counts

    in State newOptions newCounts

adjustAffectedCounts :: Sudoku (Maybe (Set Field)) -> Sudoku (Maybe (Set Field)) -> Set FieldIndex -> Map Int (Set FieldIndex) -> Map Int (Set FieldIndex)
adjustAffectedCounts newOptions oldOptions affectedFields counts = Set.foldl' (adjustCount newOptions oldOptions) counts affectedFields

adjustCount :: Sudoku (Maybe (Set Field)) -> Sudoku (Maybe (Set Field)) -> Map Int (Set FieldIndex) -> FieldIndex -> Map Int (Set FieldIndex)
adjustCount oldOptions newOptions counts position = let
  oldOptionCount = maybe 0 Set.size $ Representable.index oldOptions position
  newOptionCount = maybe 0 Set.size $ Representable.index newOptions position

  in Map.adjust (Set.delete position) oldOptionCount
    >>> Map.adjust (Set.insert position) newOptionCount
    $ counts

adjustAffectedOptions :: Set FieldIndex -> Field -> Sudoku (Maybe (Set Field)) -> Sudoku (Maybe (Set Field))
adjustAffectedOptions fields value = imapRep adjustOptionField
  where
    adjustOptionField i
      | i `Set.member` fields = fmap (Set.delete value)
      | otherwise = id
-- Set.foldl' (removeOption value) options fields: Folding took a lot of time because of immutability


